import base64
from pathlib import Path
from typing import Union

import requests
from requests import Response

from .graph import Graph


class Mermaid:
    def __init__(self, graph: Graph):
        self._diagram = self._process_diagram(graph.script)
        self._make_request_to_mermaid()

    @staticmethod
    def _process_diagram(diagram: str) -> str:
        graphbytes = diagram.encode('utf8')
        base64_bytes = base64.b64encode(graphbytes)
        diagram = base64_bytes.decode('ascii')
        return diagram

    def _repr_html_(self) -> str:
        return self.svg_response.text

    def _make_request_to_mermaid(self) -> None:
        self.svg_response: Response = requests.get('https://mermaid.ink/svg/' +
                                                   self._diagram)
        self.img_response: Response = requests.get('https://mermaid.ink/img/' +
                                                   self._diagram)

    def to_svg(self, path: Union[str, Path]) -> None:
        with open(path, 'w', encoding='utf-8') as file:
            file.write(self.svg_response.text)

    def to_png(self, path: Union[str, Path]) -> None:
        with open(path, 'wb') as file:
            file.write(self.img_response.content)



from typing import Optional, Union


class Entity:
    def __init__(self,
                 name: str,
                 attributes: dict[str, Union[list[str], str]] = None) -> None:
        self.name: str = name
        self.attributes: dict[str, Union[
            list[str], str]] = attributes if attributes is not None else {}

    def __str__(self) -> str:
        string_attributes: str = self.build_attributes()
        string: str = ''.join([self.name, '{\n', string_attributes, '}'])
        return string

    def update_attributes(self, attributes: dict[str, str]) -> None:
        self.attributes.update(attributes)

    def add_attribute(self,
                      name: str,
                      type_: str,
                      constrint: Optional[str] = None,
                      comment: Optional[str] = None) -> None:
        attribute_def: list[str] = [type_]
        if constrint is not None:
            attribute_def.append(constrint)
        if comment is not None:
            attribute_def.append(comment)
        self.attributes[name] = attribute_def

    def build_attributes(self) -> str:
        string_attributes: str = ''
        for att_name, att_def in self.attributes.items():
            string: str = ''
            if isinstance(att_def, str):
                string = f'\t{att_def} {att_name}\n'
            else:
                if len(att_def) == 1:
                    string = f'{att_def[0]} {att_name}\n'
                elif len(att_def) == 2:
                    if att_def[1] in ['PK', 'FK', 'UK']:
                        string = f'\t{att_def[0]} {att_name} {att_def[1]}\n'
                    else:
                        string = f'\t{att_def[0]} {att_name} "{att_def[1]}"\n'
                else:
                    string = f'\t{att_def[0]} {att_name} {att_def[1]} "{att_def[2]}"\n'
            string_attributes += string
        return string_attributes



from .node import Node

LINK_SHAPES: dict[str, str] = {
    'normal': '--',
    'dotted': '-.-',
    'thick': '==',
    'hiden': '~~~'
}

LINK_HEADS: dict[str, str] = {
    'none': '',
    'arrow': '>',
    'left-arrow': '<',
    'bullet': 'o',
    'cross': 'x',
}


class Link:
    def __init__(self,
                 origin: Node,
                 end: Node,
                 shape: str = 'normal',
                 head_left: str = 'none',
                 head_right: str = 'arrow',
                 message: str = '') -> None:
        self.oigin: Node = origin
        self.end: Node = end
        self.head_left = LINK_HEADS[head_left]
        self.head_right = LINK_HEADS[head_right]
        self.shape = LINK_SHAPES[shape]
        self.message = f'|{message}|' if message else message

    def __str__(self) -> str:
        if self.message:
            element: list[str] = [
                self.oigin.id_,
                ' ',
                self.head_left,
                self.shape,
                self.head_right,
                self.message,
                ' ',
                self.end.id_,
            ]
        else:
            element = [
                self.oigin.id_,
                ' ',
                self.head_left,
                self.shape,
                self.head_right,
                ' ',
                self.end.id_,
            ]
        return ''.join(element)



import os
from dataclasses import dataclass
from pathlib import Path
from typing import Optional


@dataclass
class Graph:
    """Graph is Base Class witch the other class inhirite from like `flowcahrt` , `erDiagram` , etc..

    Graph has two attribure:
    - title: string represent the title of the digrame.
    - script: string that store the main script to create the digram.

    examples:
    >>> from mermaid.graph import Graph
    >>> grph = Graph("title","...")
    >>> # save the script in txt file for other uses
    >>> graph.save() # save in file with name of title and extention .txt.
    >>> graph.svae('./file-name.txt') # save in specific file.
    """
    title: str
    script: str

    def save(self, path: Optional[Path] = None) -> None:
        file_path: Path = path if path else Path(f'./{self.title}.txt')

        with open(file_path, 'w') as file:
            file.write(self.script)

    def _build_script(self) -> None:
        script: str = f'---\ntitle: {self.title}\n---' + self.script
        self.script = script



from .entity import Entity

LIST_CARDINALITIES: dict[str, list[str]] = {
    'zero-or-one': ['|o', 'o|'],
    'exactly-one': ['||', '||'],
    'zero-or-more': ['}o', 'o{'],
    'one-or-more': ['}|', '|{'],
}


class Link:
    def __init__(self,
                 origin: Entity,
                 end: Entity,
                 origin_cardinality: str,
                 end_cardinality: str,
                 label: str = '',
                 dotted: bool = False) -> None:
        self.origin: Entity = origin
        self.end: Entity = end
        self.label: str = label
        self.left_symbol: str = LIST_CARDINALITIES[origin_cardinality][0]
        self.right_symbol: str = LIST_CARDINALITIES[end_cardinality][1]
        self.dotted: bool = dotted

    def __str__(self) -> str:
        shape: str = '..' if self.dotted else '--'
        return f'{self.origin.name}{self.left_symbol}{shape}{self.right_symbol}{self.end.name} : "{self.label}"'



from mermaid.graph import Graph

from .link import Link
from .node import Node


class FlowChart(Graph):
    def __init__(self,
                 title: str,
                 nodes: list[Node] = None,
                 links: list[Link] = None,
                 orientation: str = 'TB') -> None:
        super().__init__(title, '')
        self.orientation: str = orientation
        self.nodes: list[Node] = nodes if nodes is not None else []
        self.links: list[Link] = links if links is not None else []
        self._build_script()

    def _build_script(self) -> None:
        super()._build_script()
        script: str = f'\nflowchart {self.orientation}'
        for node in self.nodes:
            script += f'\n\t{node}'
        for link in self.links:
            script += f'\n\t{link}'
        self.script += script + '\n'



from mermaid.graph import Graph


class PieChart(Graph):
    def __init__(self,
                 title: str,
                 data: dict[str, float],
                 show_data: bool = False):
        super().__init__(title, '')
        self.data: dict[str, float] = data
        self.show_data: bool = show_data
        self._build_script()

    def _build_script(self) -> None:
        super()._build_script()
        print(self.title)
        print(self.script)
        showData: str = ' showData' if self.show_data else ''
        script: str = f'\npie{showData}'

        for key, value in self.data.items():
            script += f'\n\t"{key}" : {value}'

        self.script += script + '\n'



from dataclasses import dataclass

from mermaid import text_to_snake_case


@dataclass
class NodeShape:
    start: str
    end: str


NODE_SHAPES: dict[str, NodeShape] = {
    'normal': NodeShape('[', ']'),
    'round-edge': NodeShape('(', ')'),
    'stadium-shape': NodeShape('([', '])'),
    'subroutine-shape': NodeShape('[[', ']]'),
    'cylindrical': NodeShape('[(', ')]'),
    'circle': NodeShape('((', '))'),
    'label-shape': NodeShape('>', ']'),
    'rhombus': NodeShape('{', '}'),
    'hexagon': NodeShape('{{', '}}'),
    'parallelogram': NodeShape('[/', '/]'),
    'parallelogram-alt': NodeShape('[\\', '\\]'),
    'trapezoid': NodeShape('[/', '\\]'),
    'trapezoid-alt': NodeShape('[\\', '/]'),
    'double-circle': NodeShape('(((', ')))'),
}

HREF_TYPES: dict[str, str] = {
    'blank': '_blank',
    'self': '_self',
    'parent': '_parent',
    'top': '_top'
}


class Node:
    def __init__(self,
                 id_: str,
                 content: str = '',
                 shape: str = 'normal',
                 sub_nodes: list['Node'] = None,
                 href: str = None,
                 href_type: str = 'blank') -> None:

        self.id_: str = text_to_snake_case(id_)
        self.content: str = content if content else id_
        self.shape: NodeShape = NODE_SHAPES[shape]
        self.href: str = href if href is not None else '#'
        self.href_type: str = HREF_TYPES[href_type]
        self.sub_nodes: list[
            'Node'] = sub_nodes if sub_nodes is not None else []

    def __str__(self) -> str:
        string: str = ''
        if len(self.sub_nodes):
            string = '\n'.join([
                f'subgraph {self.id_} ["{self.content}"]',
                '\n'.join([str(f'\t{node}') for node in self.sub_nodes]), 'end'
            ])
        else:
            string = ''.join([
                self.id_, self.shape.start, f'"{self.content}"', self.shape.end
            ])
            if self.href != '#':
                string = ''.join([
                    string, '\n',
                    f'click {self.id_} "{self.href}" {self.href_type}'
                ])
        return string



import os
import re
from pathlib import Path
from typing import Optional

from .graph import Graph


def load(path: Path) -> Graph:
    script: Optional[str] = None
    name: Optional[str] = None

    if os.path.exists(path):
        with open(path, 'r') as file:
            script = file.read()
    else:
        raise FileNotFoundError(f'could not find a file in path::{path}')

    file_name: str = os.path.basename(path)
    name, _ = os.path.splitext(file_name)

    return Graph(name, script)


def text_to_snake_case(text: str) -> str:
    # Remove non-alphanumeric characters except underscores and replace spaces with underscores
    out: str = re.sub(r'[^a-zA-Z0-9_]', '_', text)

    # Convert the text to lowercase
    out = out.lower()

    return out



from mermaid.graph import Graph

from .entity import Entity
from .link import Link


class ERDiagram(Graph):
    def __init__(self,
                 title: str,
                 entities: list[Entity] = None,
                 links: list[Link] = None) -> None:
        super().__init__(title, '')
        self.entities: list[Entity] = entities if entities is not None else []
        self.links: list[Link] = links if links is not None else []
        self._build_script()

    def _build_script(self) -> None:
        super()._build_script()
        script: str = '\nerDiagram'
        for entity in self.entities:
            script += f'\n\t{entity}'
        for link in self.links:
            script += f'\n\t{link}'
        self.script += script + '\n'
